#include "locserver.h"

static const int verbose_ = 1;

// Returns a random number between 0 and max, implemented by dmaltz
static inline double
jitter (double max, int be_random_)
{
    return (be_random_ ? Random::uniform(max) : 0);
}

static bool
BackOffTest(HGPSEntry *e, double time)
    // look at the entry and decide if we can send another route
    // request or not.  update entry as well, by dmaltz
{
    double next = ((double) (0x1<<(e->rt_reqs_outstanding*2))) * RT_RQ_PERIOD;
    if (next> RT_RQ_MAX_PERIOD) next = RT_RQ_MAX_PERIOD;
    if (next + e->last_rt_req > time) return false;
    // don't let rt_reqs_outstanding overflow next on the LogicalShiftsLeft's
    if (e->rt_reqs_outstanding < 15) e->rt_reqs_outstanding++;
    e->last_rt_req = time;
    return true;
}

/* test if src and dst are in the same grid. At least src or dst must be
   a higher level grid in order for the function to be true*/
static bool
inSameGrid(location src, location dst)
{
     int m_mask;
     int mask;
     
     if (src.mask > dst.mask) {
	 m_mask = dst.mask;
     }else{
	 m_mask = src.mask;
     }
    
     mask = 0xffffffff << (BIGGEST_MASK - m_mask);
     return (((src.loc & mask) == (dst.loc & mask))? true:false);
}

/* am "I" closer to "dst" than "he" is to "dst" in ID space? */
static bool
closer(nsaddr_t me, nsaddr_t him, nsaddr_t dst)
{
    int my_dist = me - dst;
    int his_dist = him - dst;

    if (me == him) return false;

    if (my_dist == 0) return true;
    if (his_dist == 0) return false;

    if (((my_dist > 0) && (his_dist<0))||
	((my_dist < 0) && (his_dist<0) && (my_dist < his_dist)) ||
	((my_dist > 0) && (his_dist>0) && (my_dist < his_dist))) {
	return true;
    }else{
	return false;
    }
}

/* this function is for predicting how soon the node will move out of the
   range defined by min-max given its current speed and position.
   the speed is given in speed along x-axis or y-axis*/
static inline double
getOutofRangeTime(double min, double max, double speed, double now_pos)
{
    double when;
    if (speed>0) {
	when = (double)((max-now_pos)/(speed));	
    }else{
	when = (double)((now_pos-min)/(-speed));
    }
    assert(when>0);
    return when;

}

/* given a grid location, this function calculates the corresponding x, y
   coordinate range */
static inline void
getXY(double &min_x, double &max_x, double &min_y,double &max_y, location loc)
{
    int stop =loc.mask;
    int shift = BIGGEST_MASK;
    unsigned int tmp_mask = 0xffffffff;
    int the_loc = loc.loc;
    int x,y;

    x = 0;
    y = 0;
    while (stop>=2) {
	shift = shift -2;
	x = x *2 + 2* (the_loc >> (shift+1));
	y = y *2 + 2* ((the_loc & 0x55555555)>>shift);
	stop -= 2;
	//clear the examined two bits
	tmp_mask = tmp_mask >> 2;
	the_loc = the_loc & tmp_mask;
    }
    if (shift == 0) {
	x = x +1;
	y = y +1;
    }else {
	x = x*((int)pow(2,(shift/2)));
	x = x+((int)pow(2,(shift/2)));
	y = y*((int)pow(2,(shift/2)));
	y = y+((int)pow(2,(shift/2)));
    }
  
    int minus = BIGGEST_MASK - loc.mask;
    minus = minus/2;
    minus = (int)pow(2,minus);
    min_x = (x - minus)* 125;
    max_x = (x + minus) * 125;
    min_y = (y - minus) * 125;
    max_y = (y + minus) * 125;

}
/* LocServerTriggerHandler is actually a periodic handler, it is invoked
   every TRIGGER_PERIOD seconds to check if any loc updates need to be sent*/
LocServerTriggerHandler::LocServerTriggerHandler(LocServer *s, HGPSAgent *p)
{
    prev_location.loc = 0;
    prev_location.mask = 0;

    locserver = s;
    parent = p;

    /* log the loc table size later */
    counter = -4;
}

void
LocServerTriggerHandler::handle(Event *e)
{
    location now_location; 
    
    now_location = locserver->getGrid();
   
    if ((prev_location.loc == 0) && (prev_location.mask == 0)) {
	prev_location = now_location;
	locserver->myloc = now_location;

#ifdef TRIGGER_UPDATE_ONLY
	//the node has just started, send updates to all levels
	locserver->updateLoc(locserver->max_mask);
#endif
    }else {	
	
	prev_location = locserver->myloc;
	// update locserver's myloc
	locserver->myloc = now_location;

//#ifdef TRIGGER_UPDATE_ONLY
	/*check how much off I am in the hierarchical level from
	my previous position
	*/
	int ud_hierarchy_level = locserver->max_mask;
	//same_mask = (0xffffffff)>>(BIGGEST_MASK - ud_hierarchy_level);
	int same_mask = (3) << (ud_hierarchy_level-2);
	while (ud_hierarchy_level) {
	    if ((now_location.loc & same_mask) != (prev_location.loc & same_mask)) {
		locserver->updateLoc(ud_hierarchy_level); //update my old servers
		break;
	    }
	    ud_hierarchy_level -= 2;
	    same_mask = same_mask >> 2;
	}
//#endif

#ifndef TRIGGER_UPDATE_ONLY
	//spread out forwarding pointers
	if (now_location.loc != prev_location.loc) {
	    //update people in the previous grid
	    //parent->trace("HGPS trigger update _%d_ %.2f (now %d, prev%d)",parent->myaddr_,Scheduler::instance().clock(),now_location.loc,prev_location.loc);
	    locserver->updateFPGrid(prev_location);
	    /* discard all my previous fp table */
	    locserver->dumpFPTable();    
	}
	//periodic update of location info.
	locserver->checkPeriodicUpdateLoc();
#endif

    }

    //log my loctable size every 10*TRIGGER_PERIOD seconds
    counter++;
    if (counter == 10) {
	locserver->logTableSize();
	counter = 0;
    }

#ifdef TRIGGER_UPDATE_ONLY
    /* some updates may be expiring prematurely due to node changing 
       movement destinations*/
    locserver->checkExpiringUpdates();
#endif

        //reschedule the next timer event
    Scheduler::instance().schedule(this,e,TRIGGER_PERIOD);
    
}


LocServer::LocServer(HGPSAgent *p) : request_table(128) {

    parent = p;
    
    loc_trigger_handler = new LocServerTriggerHandler(this, parent);
    trigger_event = new Event();

    table = new LocTable(250,false);
    loc_cache = new LocTable(100,true);
    live_connections = new LocTable(10,true);

    /* initialize all forwarding pointer table */
    for (int i=0;i<MAX_FP_TABLE_SIZE;i++) {
	fp_table[i].id = 0;
    }
    fp_table_num = 0;
    fp_table_tail = 0;
}

void
LocServer::setRouter(HGPSRouter *r)
{
    router = r;
}

void
LocServer::setMobileNode(MobileNode *m)
{
    mynode = m;
}

/* start locserver's timer event, begin location update */
void
LocServer::start() {

    //get the max x, y coordinates of the entire terrain
    double maxx = mynode->T->upperX();
    double maxy = mynode->T->upperY();

    double len;

    if (maxx>maxy) {
	len = maxx;
    }else{
	len = maxy;
    }

    len = len - 0.1;
    max_mask = 0;
    max_len = SMALLEST_GRID;

    while (max_len < len) {
	max_len = max_len * 2;
	max_mask = max_mask + 2;
    }

 
    //prepare for the periodic update of my location 
    prev_x = -50;
    prev_y = -50;
    prev_level = max_mask-2;
    prev_timeout= -50;

    //makes sure when periodic update checker first starts, it will update all levels
    prev_level = (int)pow(2,max_mask/2) - 2;

    Scheduler::instance().schedule(loc_trigger_handler, trigger_event,5+jitter(HGPS_LOC_STARTUP_JITTER,1));

    expiration_times = new double[max_mask/2];
}

/* put locserver to sleep, could be used to emulate node crash */
void
LocServer::sleep()
{
    Scheduler::instance().cancel(trigger_event);
    table->cleanTable();
    loc_cache->cleanTable();
    live_connections->cleanTable();
}

/* get the grid position of myself*/
location
LocServer::getGrid() {
  double x,y,z;

  mynode->getLoc(&x,&y,&z);
  return getGrid(x,y);
}

double
LocServer::getSpeed() {
    return mynode->getSpeed();
}

location
LocServer::getGridPos(double &my_x, double &my_y) {
    double x, y, z;
    mynode->getLoc(&x,&y,&z);
    my_x = x;
    my_y = y;
    return getGrid(x,y);
}

void
LocServer::getPos(double &my_x, double &my_y) {
    double x, y,z;
    mynode->getLoc(&x,&y,&z);
    my_x = x;
    my_y = y;
}

/* get someone's position grid given x and y */
location
LocServer::getGrid(double x, double y)
{
     double len_x;
     double len_y;
     int grid_num = 0;
     location loc;
     
     /* assume minx = miny = 0. otherwise, use x - minx in place of x 
	similar for y
     */
     len_x = len_y = max_len;

     while ((len_x>SMALLEST_GRID)|| (len_y > SMALLEST_GRID)) {
       if (x < len_x/2) {
	 if (y < len_y/2) {
	   grid_num = grid_num * 4 + 0;
	 }else{
	   y -= len_y/2;
	   grid_num = grid_num * 4 + 1;
	 }
       }else{
	 x -= len_x/2;
	 if (y<len_y/2) {
	   grid_num = grid_num * 4 + 2;
	 }else{
	   y -= len_y/2;
	   grid_num = grid_num * 4 + 3;
	 }
       }
       len_x = len_x/2;
       len_y = len_y/2;
     }

     loc.loc = grid_num;
     loc.mask = BIGGEST_MASK;
     return loc;
}

/* hear forwarding poniter information from other's gossips, or from the 
   guy who has left. */
void
LocServer::recvFPUpdate(Packet *packet)
{
     hdr_hgps *hgpsh =  (hdr_hgps*)packet->access(parent->off_hgps_);
    
     if (myloc.loc != hgpsh->fp_old_loc.loc) {
	//i don't belong to the grid that should contain the FP pointers
	return;
     }

     double oldest_lastupdate_time = Scheduler::instance().clock();
     int oldest_entry = -1;

     for (int i=0;i<hgpsh->fp_num_;i++) {
	/*i maintain a fixed array for FP pointer table;
	  If I run out of space, I'll just get rid of the oldest one. */
	 int j;
	 for (j=0;j<MAX_FP_TABLE_SIZE;j++) {
	     if (fp_table[j].id == hgpsh->fp_ids_[i]) {
		 if (fp_table[j].lastupdate_time <= hgpsh->fp_lastupdate_time_[i]) {
		     fp_table[j].loc = hgpsh->fp_locs_[i];
		 }
		 break;
	     }else if (fp_table[j].id == 0) {
		 fp_table_num++; //cause I add a new entry
		 assert(fp_table_num <= MAX_FP_TABLE_SIZE);
		 break;
	     }

	     if (fp_table[j].lastupdate_time < oldest_lastupdate_time) {
		 oldest_entry = j;
		 oldest_lastupdate_time = fp_table[j].lastupdate_time;
	     }
	 }
	 
	 if (j==MAX_FP_TABLE_SIZE) {
	     //i.e. there's no existing fp pointer for this id, add a new one
	     fp_table[oldest_entry].id = hgpsh->fp_ids_[i];
	     fp_table[oldest_entry].lastupdate_time = hgpsh->fp_lastupdate_time_[i];
	     fp_table[oldest_entry].loc = hgpsh->fp_locs_[i];
	 }else {
	     fp_table[j].id = hgpsh->fp_ids_[i];
	     fp_table[j].lastupdate_time = hgpsh->fp_lastupdate_time_[i];
	     fp_table[j].loc = hgpsh->fp_locs_[i];
	 }

    }
    
}

/* gossip about FP pointers for the nodes in my grid 
   serveral FP pointers will be randomly selected and 
   piggybacked onto the packet passed in as argument
   this function is to be called by routing layer*/
void
LocServer::piggybackFP(Packet *packet)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)packet->access(parent->off_hgps_);
    int which_fp;

    if (fp_table_num <= MAX_FP_IN_HELLO) {
	for (int i=0;i<fp_table_num;i++) {
	    hgpsh->fp_ids_[i] = fp_table[i].id;
	    hgpsh->fp_locs_[i] = fp_table[i].loc;
	    hgpsh->fp_lastupdate_time_[i] = fp_table[i].lastupdate_time;
	}
	hgpsh->fp_num_ = fp_table_num;
	hgpsh->fp_old_loc = myloc;

    }else {

	for (int i=0;i<MAX_FP_IN_HELLO;i++) {
	    which_fp = (int)(Random::uniform()*fp_table_num);
	    hgpsh->fp_ids_[i] = fp_table[which_fp].id;
	    hgpsh->fp_locs_[i] = fp_table[which_fp].loc;
	    hgpsh->fp_lastupdate_time_[i] = fp_table[which_fp].lastupdate_time;
	}
	hgpsh->fp_num_ = MAX_FP_IN_HELLO;
	hgpsh->fp_old_loc = myloc;
    }
}

void
LocServer::dumpFPTable()
{
    for (int i=0;i<MAX_FP_TABLE_SIZE;i++) {
	fp_table[i].id = 0;
    }
    fp_table_num = 0;
    fp_table_tail = 0;
}

#ifdef TRIGGER_UPDATE_ONLY
void 
LocServer::recvUpdate(Packet *packet)
{
      hdr_hgps *hgpsh =  (hdr_hgps*)packet->access(parent->off_hgps_);
      hdr_ip *iph =  (hdr_ip*)packet->access(parent->off_ip_);
      loc_entry closest;
      assert(hgpsh->type_ == LOC_UPDATE);

      int divide = (BIGGEST_MASK - hgpsh->update_loc_.mask + 2)/2;
      loc_cache->addUpdateEntry(iph->src(),hgpsh->src_loc_,(double)(hgpsh->timeout_inteval_/divide));

      /* if I'm not in the update area, forward the packet on*/
      //if ((!inSameGrid(myloc,hgpsh->update_loc_)) && (!inSameGrid(hgpsh->dst_loc_, hgpsh->update_loc_))) {
      if ((!inSameGrid(myloc,hgpsh->update_loc_))) {

	  if (hgpsh->in_correct_grid_ && (iph->dst()!=parent->myaddr_)) {
	      //This packet has been in the correct grid before, probably the previous guy just asks me to relay the packet
	      router->sendOutPktWithLocation(packet,0.0,true);
	      return;
	  }

	  /* if somebody mistook me as being in the to-be-updated grid, or
	     the update packet has been in the to-be-updated grid before, 
	     i should try to forward to somebody I know that's closer to 
	     dst in ID space than me.  and who is actually in that to-be-updated grid*/
	  if (hgpsh->in_correct_grid_) {
	  }else if ((iph->dst() == parent->myaddr_) && ((hgpsh->dst_loc_.mask == BIGGEST_MASK) && inSameGrid(hgpsh->dst_loc_,hgpsh->update_loc_))) {
	    hgps->in_correct_grid_ = 1;
	  }else{
	    hgpsh->dst_loc_ = hgpsh->update_loc_;
	    iph->dst() = 0;
	    parent->trace("HGPS: _%d_(grid %d) forward update(%d) to correct grid %d",parent->myaddr_,myloc.loc,iph->src(),hgpsh->update_loc_.loc);
	    router->sendOutPktWithLocation(packet,0.0,true);
	    return;
	  }
      }else {
	  if (!hgpsh->in_correct_grid_) {
	    hgpsh->in_correct_grid_ = 1;
	    iph->dst() = 0;
	  }
      }

      /* see if I have to forward the packet on or should I just update my own location table*/
      if (hgpsh->next_locserver_ != 0) {
	  closest.id = hgpsh->locservers_[hgpsh->next_locserver_-1].id;
	  closest.loc = hgpsh->locservers_[hgpsh->next_locserver_-1].loc;
      }else{
	  closest.id = 0;
      }

      table->findClosestExceptDst(parent->myaddr_,myloc,iph->src(),router->nt,hgpsh->update_loc_,closest);

      if ( closest.id == parent->myaddr_) {

	  //I am the closest node to the updater, i should keep his information
	  if (parent->verbose_) {
	      parent->trace("HGPS: _%d_(%d) add (%d,loc %d) update area(%d,%d) to table %.5f",parent->myaddr_,
		      myloc.loc,iph->src(),hgpsh->src_loc_.loc,hgpsh->update_loc_.loc,hgpsh->update_loc_.mask,(Scheduler::instance().clock()+hgpsh->timeout_interval_));
	  }
	  hgpsh->src_loc_.mask = hgpsh->update_loc_.mask;
	  table->addUpdateEntry(iph->src(),hgpsh->src_loc_,hgpsh->timeout_inteval_);
	  Packet::free(packet);
      
      }else if ((!iph->dst()) || (closest.id != hgpsh->locservers_[hgpsh->next_locserver_-1].id)) {
	  
	  hgpsh->locservers_[0].id = closest.id;
	  hgpsh->locservers_[0].loc = closest.loc;
	  hgpsh->next_locserver_ = 1;
	  iph->dst() = closest.id;
	  hgpsh->dst_loc_ = closest.loc;
	  //hgpsh->lastupdate_dst_time_ = closest.lastupdate_time;
	  hgpsh->prev_hop_num_ = 0;

	  router->sendOutPktWithLocation(packet,0.0,true);
	  
      }else { //closest.id == hgpsh->locservers_[hgpsh->next_locserver-1].id
	 
	  if (iph->dst() == parent->myaddr_) {
	      assert(hgpsh->locservers_[hgpsh->next_locserver_ - 1].id == parent->myaddr_);	
	      hgpsh->next_locserver_--;
	      if (hgpsh->next_locserver_ == 0) {
		  //location query failure
		  if (parent->verbose_) {
		      parent->trace("HGPS %.5f _%d_ loc update location failure %d(%d,%d)",Scheduler::instance().clock(),parent->myaddr_,iph->src(),hgpsh->update_loc_.loc,hgpsh->update_loc_.mask);
		  }
		  Packet::free(packet);
		  return;
	      }
	      iph->dst() = hgpsh->locservers_[hgpsh->next_locserver_ - 1].id;
	      hgpsh->dst_loc_ = hgpsh->locservers_[hgpsh->next_locserver_-1].loc;
	      assert(inSameGrid(myloc,hgpsh->dst_loc_));
	  }
	  //check if I am in the same grid as the destination grid and need to restart query
	  if (inSameGrid(myloc, hgpsh->dst_loc_) && (hgpsh->dst_loc_.mask != BIGGEST_MASK)) {
	      //I have to query for iph->dst in order to continue
	      bool exists = table->findClosestInGrid(parent->myaddr_,myloc,iph->dst(),hgpsh->dst_loc_,router->nt,closest,0.0); //jinyang -- I'm not using hgpsh->lastupdate_dst_time_, worth more consideration
	      if (!exists) {
		  //location query failure
		  parent->trace("HGPS %.5f _%d_(%d) update loc failure %d(%d %d)->%d",Scheduler::instance().clock(),parent->myaddr_,myloc.loc,iph->src(),hgpsh->update_loc_.loc,hgpsh->update_loc_.mask,iph->dst());
		  Packet::free(packet);
		  return;
	      }
	      if (parent->verbose_) {
		  parent->trace("HGPS _%d_ (update myloc %d) recursive query for next location server %d (%d,%d), find server %d(%d,%d)", parent->myaddr_,myloc.loc,iph->dst(),hgpsh->dst_loc_.loc,hgpsh->dst_loc_.mask,closest.id,closest.loc.loc,closest.loc.mask);
	      }


	      //put iph->dst() in hgpsh->locservers_
	      assert(hgpsh->next_locserver_<MAX_LOC_SERVERS);
	      if (iph->dst()!=closest.id) {
		  hgpsh->locservers_[hgpsh->next_locserver_].loc = closest.loc;
		  hgpsh->locservers_[hgpsh->next_locserver_].id = closest.id;
		  hgpsh->next_locserver_++;
	      }

	      iph->dst() = closest.id;
	      hgpsh->dst_loc_ = closest.loc;
	      //hgpsh->lastupdate_dst_time_ = closest.lastupdate_time;
	      hgpsh->prev_hop_num_ = 0;
	  }

	  router->sendOutPktWithLocation(packet,0.0,true);
      }
}

#else

void 
LocServer::recvUpdate(Packet *packet)
{
      hdr_hgps *hgpsh =  (hdr_hgps*)packet->access(parent->off_hgps_);
      hdr_ip *iph =  (hdr_ip*)packet->access(parent->off_ip_);
      loc_entry closest;
      assert(hgpsh->type_ == LOC_UPDATE);

      int divide = (BIGGEST_MASK - hgpsh->update_loc_.mask + 2)/2;
      loc_cache->addUpdateEntry(iph->src(),hgpsh->src_loc_,(double)(hgpsh->timeout_inteval_/divide));


      /* if I'm not in the update area, forward the packet on*/
      if ((!inSameGrid(myloc,hgpsh->update_loc_))) {

	  if (hgpsh->in_correct_grid_ && (iph->dst()!=parent->myaddr_)) {
	      //This packet has been in the correct grid before, probably the previous guy just asks me to relay the packet
	      parent->trace("HGPS: _%d_ forward directly to %d(%d) in update area (%d,%d)",parent->myaddr_,iph->dst(),hgpsh->dst_loc_.loc,hgpsh->update_loc_.loc,hgpsh->update_loc_.mask);
	      router->sendOutPktWithLocation(packet,0.0,true);
	      return;
	  }

	  if (hgpsh->in_correct_grid_) {
	      
           /* if somebody mistook me as being in the to-be-updated grid, or
	     the update packet has been in the to-be-updated grid before, 
	     i should try to forward to somebody I know that's closer to 
	     dst in ID space than me.  and who is actually in that to-be-updated grid*/
	  }else if ((iph->dst() == parent->myaddr_) && ((hgpsh->dst_loc_.mask == BIGGEST_MASK) && inSameGrid(hgpsh->dst_loc_,hgpsh->update_loc_))) {
	    hgpsh->in_correct_grid_ = 1;
	  }else{
	    hgpsh->dst_loc_ = hgpsh->update_loc_;
	    iph->dst() = 0;
	    parent->trace("HGPS: _%d_(grid %d) forward update(%d) to correct grid %d",parent->myaddr_,myloc.loc,iph->src(),hgpsh->update_loc_.loc);
	    router->sendOutPktWithLocation(packet,0.0,true);
	    return;
	  }
      }else {
	  hgpsh->in_correct_grid_ = 1;
      }

      /* see if I have to forward the packet on or should I just update my own location table*/
      if (iph->dst()) {
	  closest.id = iph->dst();
	  closest.loc = hgpsh->dst_loc_;
      }

      table->findClosestExceptDst(parent->myaddr_,myloc,iph->src(),router->nt,hgpsh->update_loc_,closest);

      if ( closest.id == parent->myaddr_) {
	  //I am the closest node to the updater, i should keep his information
	  if (parent->verbose_) {
	      parent->trace("HGPS: _%d_(%d) add (%d,loc %d) update area(%d,%d) to table %.5f",parent->myaddr_,
		      myloc.loc,iph->src(),hgpsh->src_loc_.loc,hgpsh->update_loc_.loc,hgpsh->update_loc_.mask,(Scheduler::instance().clock()+hgpsh->timeout_inteval_));
	  }
	  table->addUpdateEntry(iph->src(),hgpsh->src_loc_,hgpsh->timeout_inteval_);
	  Packet::free(packet);
      }else {
	  hgpsh->dst_loc_ = closest.loc;
	  //I should forward the packet to the closest next guy
	  if (parent->verbose_) {
	      parent->trace("HGPS: _%d_(%d) forward to %d(%d,%d) nearer to (%d)",parent->myaddr_,myloc.loc,closest.id,closest.loc.loc,closest.loc.mask,iph->src());
	  }
	  if (iph->dst() != closest.id) {
	    hgpsh->prev_hop_num_ = 0;
	    iph->dst() = closest.id;
	   }
	  router->sendOutPktWithLocation(packet,0.0,true);
      }
}
#endif

/* hgpsagent calls this function to obtain the location information for dst
   it returns immmediately if dst is in loc_cache or loc_table
   otherwise, a LOC_QUERY is sent out in search of dst's location server */

bool
LocServer::getLocation(nsaddr_t dst, location & dst_loc)
{

    double now = Scheduler::instance().clock();
    loc_entry closest;
    bool exists = false;

    dst_loc.loc = -1;
    dst_loc.mask = BIGGEST_MASK;

    //try to get location from location cache
    loc_entry *entry = loc_cache->getEntry(dst,now);

    if (entry && ((now - entry->lastupdate_time)<CACHE_VALID_TIME)) { 
	dst_loc = entry->loc;
	if (parent->verbose_) {
	    parent->trace("S$hit cache %.5f _%d_ ??->%d(%d)",now,parent->myaddr_,dst,dst_loc.loc);
	}

#ifndef TEST_QUERY
	/* according to the actual protocol, the location of the dst is considered known and data pkt
	   could be sent. If this program is used for testing accuracy of GLS
	   query, a query packet is directly sent to the dst to check that the location is
	   indeed valid. */
	return true;
#else
        closest.id = dst;
        closest.loc = entry->loc;
        goto SEND_Q;
#endif
    }

    exists= table->findClosest(parent->myaddr_,myloc,dst,router->nt,closest,0.0);

    if (exists && (closest.id == dst)) {
	 if (parent->verbose_) {
	     parent->trace("S$hit loc table %.5f _%d_ ??->%d(%d)",now,parent->myaddr_,
		     dst,(closest.loc).loc);
	 }
	dst_loc = closest.loc;

#ifndef TEST_QUERY
	/* if the program is used for testing query only, a GLS query is still sent to
	   check that the location is indeed valid.*/
	return true;
#else
        goto SEND_Q;
#endif
    }

    if (!exists) {
	//in desperation, I search in the cache and 2-hop routing table. this should happen very rarely.
	exists = loc_cache->findClosest(parent->myaddr_,myloc,dst,router->nt,closest,0.0);
	if (exists && ((now-closest.lastupdate_time)<CACHE_VALID_TIME)) {
#ifdef HGPS_DEBUG
	    printf("_%d_ my cache saved this loc query ->%d\n",parent->myaddr_,dst);
#endif
	}else if ((exists = table->findClosestinRT(parent->myaddr_,myloc,dst,router->nt,closest,false))){
#ifdef HGPS_DEBUG
	    printf("_%d_ my routing table saved this loc query from %d->%d\n",parent->myaddr_,parent->myaddr_,dst);
#endif
	}else {
	    assert(parent->asleep == false);
#ifdef TEST_WS
	    /* if the program is in testing node up and down mode, I could infer some reasons
	       for the failure of location query */
	    if ((now - parent->wakeup_time) < 10) {//just some threshold value
		//it's because i have not stayed alive long enough
		parent->trace("HGPS %.5f _%d_ loc-query-failure-3(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,dst,parent->myaddr_,dst);
	    }else {
		double his_wakeuptime = (God::instance()->wakeup_list)[dst];
		if ((now-his_wakeuptime)<10) {
		    //this is because the other guy has not stayed alive long enough
		    parent->trace("HGPS %.5f _%d_ loc-query-failure-4(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,dst,parent->myaddr_,dst);
		}else {
#endif
		    printf("LOC QUERY FAILED _%d_ %d->%d\n",parent->myaddr_,parent->myaddr_,dst);
		    parent->trace("HGPS %.5f _%d_ loc-query-failure-2(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,dst,parent->myaddr_,dst);
#ifdef TEST_WS
		}
	    }
#endif
	    return false;
	}
    }

SEND_Q:

    HGPSEntry *e = request_table.getHGPSEntry(dst);

    Packet *pkt = parent->allocpkt();
    hdr_hgps *hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_);
    hdr_ip *iph =  (hdr_ip*)pkt->access(parent->off_ip_);
    hdr_cmn *cmnh =  (hdr_cmn*)pkt->access(parent->off_cmn_);
     
   
    hgpsh->init();
    hgpsh->type_ = LOC_QUERY;
    hgpsh->id_ = dst;
    hgpsh->src_loc_ = getGrid();
    
    iph->dport_ = RT_PORT;
    iph->src() = parent->myaddr_;

    cmnh->ptype() = PT_HGPS;
    cmnh->num_forwards() = 0;
    cmnh->size() = hgpsh->size();

    hgpsh->dst_loc_ = closest.loc;
    iph->dst() = closest.id;

    hgpsh->locservers_[0].id = closest.id;
    hgpsh->locservers_[0].loc = closest.loc;
    hgpsh->next_locserver_ = 1;

    if (BackOffTest(e,now)) {
	parent->trace("HGPS %.5f _%d_ new-loc-query %d->%d(%d)",now,parent->myaddr_,iph->src(),iph->dst(),dst);
	router->sendOutPktWithLocation(pkt,0.0,true,true);
    }else{
	if (parent->verbose_) {
	    parent->trace("HGPS LOC_QUERY not sent %.5f _%d_ %d(%d)",now,parent->myaddr_,iph->src(),dst);
	}
	Packet::free(pkt);
    }
    return true;
}

#ifdef TRIGGER_UPDATE_ONLY

void
LocServer::recvQuery(Packet *pkt, bool dst_loc_failure)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_);
    hdr_ip *iph =  (hdr_ip*)pkt->access(parent->off_ip_);

    loc_entry closest;
    loc_entry *cache_ent;
    bool exists;
    double now = Scheduler::instance().clock();

    /* if location failure is encountered immediately at this node, routing layer
    would call function recvQuery again with dst_loc_failure set to true */
    if (dst_loc_failure) {
        parent->trace("HGPS %.5f _%d_ loc-query-failure-1(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
	Packet::free(pkt);
	return;
    }

    /* the query target is me. Send out GLS response packet immediately */
    if (hgpsh->id_ == parent->myaddr_) {
	if (parent->verbose_) {
	    parent->trace("HGPS %.5f _%d_ loc-reply-sent %d->%d about %d(%d)", Scheduler::instance().clock(),
		parent->myaddr_, parent->myaddr_, iph->src(), hgpsh->id_, hgpsh->src_loc_);
	}
	
	//send out the GLS RESPONSE
	Packet *p = parent->allocpkt();

	hdr_hgps *new_hgpsh =  (hdr_hgps*)p->access(parent->off_hgps_); 
	hdr_ip *new_iph =  (hdr_ip*)p->access(parent->off_ip_); 
	hdr_cmn *new_cmnh =  (hdr_cmn*)p->access(parent->off_cmn_);
	new_iph->dport_ = RT_PORT;
	new_iph->src() = parent->myaddr_;

	new_hgpsh->init();
	new_hgpsh->src_loc_ = getGrid();
	new_hgpsh->type_ = LOC_QUERY_RESPONSE;

	new_hgpsh->loc_ = new_hgpsh->src_loc_;
	new_hgpsh->id_ = parent->myaddr_;
	new_iph->dst() = iph->src();
	new_hgpsh->dst_loc_ = hgpsh->src_loc_;

	//copy the variable for recording how many steps GLS query has taken
	new_hgpsh->query_forwards_ = hgpsh->query_forwards_; 

	new_cmnh->ptype() = PT_HGPS;
	new_cmnh->size() = new_hgpsh->size();

	Packet::free(pkt);
	router->sendOutPktWithLocation(p,0.0,true);
	return;
    }

    if ((iph->dst() == hgpsh->id_) && (hgpsh->dst_loc_.mask == BIGGEST_MASK)){
	//inform me of any location failures for this packet
	router->sendOutPktWithLocation(pkt,0.0,true,true);
	return;

    }else {//iph->dst()!=hgpsh->id_ , the dst location is not known yet

	//check cache for direct location entry for dst first, (just a slight optimization)
	if ((cache_ent=loc_cache->getEntry(hgpsh->id_,now))  
	    && ((now - cache_ent->lastupdate_time)<CACHE_VALID_TIME)
	    && ((iph->dst() != hgpsh->id_)||(cache_ent->lastupdate_time > hgpsh->lastupdate_time_))) {
	    
	    //good good it hits the cache!
	    parent->trace("HGPS %.2f _%d_ cache-reply-sent (%d->%d)", now, parent->myaddr_, iph->src(),hgpsh->id_);
	    closest.id = hgpsh->id_;
	    closest.loc = cache_ent->loc;
	    closest.lastupdate_time = cache_ent->lastupdate_time;
	    router->sendOutPktWithLocation(pkt,0.0,true,true);
	    return;

	} else {

	    exists = table->findClosest(parent->myaddr_,myloc,hgpsh->id_,router->nt,closest,hgpsh->lastupdate_dst_time_);
	    if (!exists) {
		goto QUERY_AGAIN;	
	    }
	}
    }
		
   //I'll have to forward the query packet on...
    if ((!iph->dst())||closer(closest.id, hgpsh->locservers_[0].id,hgpsh->id_)) {
	iph->dst() = closest.id;
	//hgpsh->lastupdate_dst_time_ = closest.lastupdate_time;
	hgpsh->dst_loc_ = closest.loc;
	hgpsh->prev_hop_num_ = 0;
	//put the loc server that is the closest to hgpsh->id_ on the head of the list, removing the old list
	hgpsh->locservers_[0].id = iph->dst();
	hgpsh->locservers_[0].loc = closest.loc;
	hgpsh->next_locserver_ = 1;

    }else if ((closest.id == iph->dst()) && ((closest.lastupdate_time-0.1) > hgpsh->lastupdate_time_)) {
	hgpsh->lastupdate_time_ = closest.lastupdate_time;
	hgpsh->dst_loc_ = closest.loc;
	//hgpsh->lastupdate_dst_time_ = closest.lastupdate_time; 
    }

QUERY_AGAIN:
    bool check_exempted = false;
    if (iph->dst() == parent->myaddr_) {
	assert(hgpsh->locservers_[hgpsh->next_locserver_ - 1].id == parent->myaddr_);	
	//assert(inSameGrid(hgpsh->locservers_[hgpsh->next_locserver_-1].loc,hgpsh->dst_loc_));
	check_exempted = true;
	hgpsh->next_locserver_--;
	if (hgpsh->next_locserver_ == 0) {
	    //location query failure
	    parent->trace("HGPS %.5f _%d_ loc-query-failure-1(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
	    Packet::free(pkt);
	    return;
	}
	iph->dst() = hgpsh->locservers_[hgpsh->next_locserver_ - 1].id;
	hgpsh->dst_loc_ = hgpsh->locservers_[hgpsh->next_locserver_-1].loc;
    }
    //check if I am in the same grid as the destination grid and need to restart query
    if ((check_exempted || inSameGrid(myloc, hgpsh->dst_loc_)) && (hgpsh->dst_loc_.mask != BIGGEST_MASK)) {
	//I have to query for iph->dst in order to continue
	exists = table->findClosestInGrid(parent->myaddr_,myloc,iph->dst(),hgpsh->dst_loc_,router->nt,closest,0.0); //jinyang hgpsh->lastupdate_dst_time_ has been deleted here
	if (parent->verbose_) {
	    parent->trace("HGPS _%d_ recursive query for next location server %d (%d,%d), find server %d(%d,%d)", parent->myaddr_,iph->dst(),hgpsh->dst_loc_.loc,hgpsh->dst_loc_.mask,closest.id,closest.loc.loc,closest.loc.mask);
	}

	if (!exists) {
	    //location query failure
	    parent->trace("HGPS %.5f _%d_ loc-query-failure-1(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
	    Packet::free(pkt);
	    return;
	}
	//put iph->dst() in hgpsh->locservers_
	assert(hgpsh->next_locserver_<MAX_LOC_SERVERS);
	if (closest.id!=iph->dst()) {
	    hgpsh->locservers_[hgpsh->next_locserver_].loc = closest.loc;
	    hgpsh->locservers_[hgpsh->next_locserver_].id = closest.id;
	    hgpsh->next_locserver_++;
	}
	iph->dst() = closest.id;
	hgpsh->dst_loc_ = closest.loc;
	//hgpsh->lastupdate_dst_time_ = closest.lastupdate_time;
    }

    router->sendOutPktWithLocation(pkt,0.0,true,true);
}

#else

/* this function handles GLS query packet */
void
LocServer::recvQuery(Packet *pkt, bool dst_loc_failure)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_);
    hdr_ip *iph =  (hdr_ip*)pkt->access(parent->off_ip_);

    loc_entry closest;
    loc_entry *cache_ent;
    bool exists;
    double now = Scheduler::instance().clock();

    /* if location failure is encountered immediately at this node, routing layer
    would call function recvQuery again with dst_loc_failure set to true */
    if (dst_loc_failure) {
	//check info in my fp_table
	for (int i=0;i<MAX_FP_TABLE_SIZE;i++) {
	   if(fp_table[i].id == iph->dst()) {
	       //my forwarding pointer saved this query
	       if ((fp_table[i].lastupdate_time > hgpsh->dst_loc_time_) && (hgpsh->dst_loc_.loc != fp_table[i].loc.loc)) {
		  hgpsh->dst_loc_ = fp_table[i].loc;
		  hgpsh->dst_loc_time_ = fp_table[i].lastupdate_time;
		  if (iph->dst() == hgpsh->id_) {
		      hgpsh->lastupdate_dst_time_ = fp_table[i].lastupdate_time;
		  }
		  if (parent->verbose_) {
		    parent->trace("HGPS %.5f _%d_ fp directs to new loc %d for loc server %d (query %d->%d)",Scheduler::instance().clock(),hgpsh->dst_loc_,iph->dst(),iph->src(),hgpsh->id_);
		  }
		  //send out the pkt, if there is error again, just discard the packet
		  router->sendOutPktWithLocation(pkt,0.0,true,false);
		  return;
	       }
	       break;
	    }
	}
	/* if location failure is encountered for the dst, and forwarding pointers
	       did not save the query, the query is restarted */
	if (iph->dst() == hgpsh->id_) {
	    table->deleteEntry(iph->dst());
	    loc_cache->deleteEntry(iph->dst());
	    iph->dst() = 0; 
	}else{
	    // forwarding pointer could not save the query, discard!
	    parent->trace("HGPS %.5f _%d_ loc-query-failure-1(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
	    Packet::free(pkt);
	    return;
	}
    }

    /* the query target is me. Send out GLS response packet immediately */
    if (hgpsh->id_ == parent->myaddr_) {
	if (parent->verbose_) {
	    parent->trace("HGPS %.5f _%d_ loc-reply-sent %d->%d about %d(%d)", Scheduler::instance().clock(),
		parent->myaddr_, parent->myaddr_, iph->src(), hgpsh->id_, hgpsh->src_loc_);
	}
	
	//send out the GLS RESPONSE
	Packet *p = parent->allocpkt();

	hdr_hgps *new_hgpsh =  (hdr_hgps*)p->access(parent->off_hgps_); 
	hdr_ip *new_iph =  (hdr_ip*)p->access(parent->off_ip_); 
	hdr_cmn *new_cmnh =  (hdr_cmn*)p->access(parent->off_cmn_);
	new_iph->dport_ = RT_PORT;
	new_iph->src() = parent->myaddr_;

	new_hgpsh->init();
	new_hgpsh->src_loc_ = getGrid();
	new_hgpsh->type_ = LOC_QUERY_RESPONSE;

	new_hgpsh->loc_ = new_hgpsh->src_loc_;
	new_hgpsh->id_ = parent->myaddr_;
	new_iph->dst() = iph->src();
	new_hgpsh->dst_loc_ = hgpsh->src_loc_;

	//copy the variable for recording how many steps GLS query has taken
	new_hgpsh->query_forwards_ = hgpsh->query_forwards_; 

	new_cmnh->ptype() = PT_HGPS;
	new_cmnh->size() = new_hgpsh->size();

	Packet::free(pkt);
	router->sendOutPktWithLocation(p,0.0,true);
	return;
    }

    if (iph->dst() == hgpsh->id_){
	/*location for the query target is known, forward GLS Query directly to dst*/
	//inform me of any location failures for this packet
	router->sendOutPktWithLocation(pkt,0.0,true,true);
	return;

    }else {//iph->dst()!=hgpsh->id_ , the dst location is not known yet

	//check cache for direct location entry for dst first, (just a slight optimization)
	if ((cache_ent=loc_cache->getEntry(hgpsh->id_,now))  
	    && ((now - cache_ent->lastupdate_time)<CACHE_VALID_TIME)
	    && ((iph->dst() != hgpsh->id_)||(cache_ent->lastupdate_time > hgpsh->lastupdate_time_))) {
	    
	    //good good it hits the cache!
	    parent->trace("HGPS %.2f _%d_ cache-reply-sent (%d->%d)", now, parent->myaddr_, iph->src(),hgpsh->id_);
	    closest.id = hgpsh->id_;
	    closest.loc = cache_ent->loc;
	    closest.lastupdate_time = cache_ent->lastupdate_time;

	} else {

	    exists = table->findClosest(parent->myaddr_,myloc,hgpsh->id_,router->nt,closest,hgpsh->lastupdate_dst_time_);
	    if (!exists) {
		if ((iph->dst()) && (closer(iph->dst(),parent->myaddr_,hgpsh->id_))) {
		    router->sendOutPktWithLocation(pkt,0.0,true,true);
		}else{
		    assert(parent->asleep == false);
		    parent->trace("HGPS %.5f _%d_ loc-query-failure-2(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
		    Packet::free(pkt);
		}
		return;
	    }
	    /*
	    if (!exists) {
		//check cache first
		exists = loc_cache->findClosest(parent->myaddr_,myloc,hgpsh->id_,router->nt,closest,hgpsh->lastupdate_dst_time_);
		if (exists && ((now-closest.lastupdate_time)<CACHE_VALID_TIME)) {
#ifdef HGPS_DEBUG
		    printf("_%d_ my cache saved this loc query from %d->%d\n",parent->myaddr_,iph->src(),hgpsh->id_);
#endif
		}else if ((exists = table->findClosestinRT(parent->myaddr_,myloc,hgpsh->id_,router->nt,closest,false))){
#ifdef HGPS_DEBUG
		    printf("_%d_ my routing table saved this loc query from %d->%d\n",parent->myaddr_,iph->src(),hgpsh->id_);
#endif
		}else {
		    //relay the packet if iph->dst() is closer to the dst than myself in ID space
		    if ((iph->dst()) && closer(iph->dst(),parent->myaddr_,hgpsh->id_)) {
			router->sendOutPktWithLocation(pkt,0.0,true,true);
		    }else{
			assert(parent->asleep == false);
#ifdef TEST_WS
			//infer some potential reasons for query failure if the program is in Test wake sleep mode
			if ((now - parent->wakeup_time) < 10) {
			    //it's because i have not stayed alive long enough
			    parent->trace("HGPS %.5f _%d_ loc-query-failure-3(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
			}else{
			    double his_wakeuptime = (God::instance()->wakeup_list)[hgpsh->id_];
			    if ((now-his_wakeuptime)<7) {
				//this is because the other guy has not stayed alive long enough
				parent->trace("HGPS %.5f _%d_ loc-query-failure-4(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,parent->myaddr_,hgpsh->id_);
			    }else{
#endif
#ifdef HGPS_DEBUG
				printf("HGPS DROP: LOC QUERY FAILED _%d_ %d->%d asking for %d\n",parent->myaddr_,iph->src(),iph->dst(), hgpsh->id_);
#endif
				parent->trace("HGPS %.5f _%d_ loc-query-failure-2(%d %d)%d->%d",now,parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
#ifdef TEST_WS
			    }
			}
#endif
			Packet::free(pkt);
		    }
		return;
		}
	    }
	    */

	}

    }

   //I'll have to forward the query packet on...
    if ((!iph->dst())||closer(closest.id, iph->dst(),hgpsh->id_)) { 
	iph->dst() = closest.id;
	if (iph->dst() == hgpsh->id_) {
	    hgpsh->lastupdate_dst_time_ = closest.lastupdate_time;
	}
	hgpsh->dst_loc_ = closest.loc;
	hgpsh->prev_hop_num_ = 0;
    }else if ((closest.id == iph->dst()) && ((closest.lastupdate_time-0.1) > hgpsh->lastupdate_time_)) { 
	hgpsh->lastupdate_time_ = closest.lastupdate_time;
	hgpsh->dst_loc_ = closest.loc;
	if (iph->dst() == hgpsh->id_) {
	    hgpsh->lastupdate_dst_time_ = closest.lastupdate_time;
	}

    }
    router->sendOutPktWithLocation(pkt,0.0,true, true);
}

#endif

/* this function is invoked when the query's initiator gets a query reply back */
void
LocServer::recvQueryResponse(Packet *pkt)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_);
 
    //i've received a query response, cache it!!
    loc_cache->addUpdateEntry(hgpsh->id_,hgpsh->loc_,hgpsh->timeout_inteval_);
    parent->trace("HGPS %.5f _%d_ loc-reply-received %d %d %d(%d)",Scheduler::instance().clock(),
	    parent->myaddr_,hgpsh->query_forwards_,hgpsh->query_response_forwards_,hgpsh->id_,hgpsh->loc_);
    parent->notifyLocation(hgpsh->id_,hgpsh->loc_);
    
    Packet::free(pkt);
}

void
LocServer::checkPeriodicUpdateLoc()
{
    double x, y, z;
    mynode->getLoc(&x,&y,&z);
    double dist = sqrt((x-prev_x)*(x-prev_x) + (y-prev_y)*(y-prev_y));
    double now = Scheduler::instance().clock();

    if ((dist >= parent->update_distance_) || (now > prev_timeout)) {
	// the second hierarchy level 's update rate is one half of the first level's etc. etc.
	prev_level = prev_level+2;
	int max =(int)pow(2,max_mask/2);
	if (prev_level > max) {
	    prev_level = 2; //reset prev_level
	}
	int level = 0;
	int tmp_prev_level = prev_level;
	while (tmp_prev_level % 2 == 0) {
	    level += 2;
	    tmp_prev_level = tmp_prev_level >> 1;
	}
	if (parent->myaddr_ == 1) {
	    printf("updating level %d %d\n",level, prev_level);
	}
	updateLoc(level);
	prev_timeout= now + (double)2 * parent->update_distance_/getSpeed();
	prev_x = x;
	prev_y = y;
    }

    Packet *pkt;
    hdr_hgps *hgpsh;
    hdr_cmn *cmnh;
    hdr_ip *iph;
    //inform the sources of active connections about location change
    if (myloc.loc != curr_loc.loc) {
	//update all my live connection peers!
	for (int i=0;i<live_connections->size;i++) {
	    if ((now-live_connections->table[i].lastupdate_time)< LIVE_CONN_TIMEOUT) {

		pkt = parent->allocpkt();

		hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_); 
		iph =  (hdr_ip*)pkt->access(parent->off_ip_); 
		cmnh =  (hdr_cmn*)pkt->access(parent->off_cmn_);

		iph->dport_ = RT_PORT;
		iph->dst() = live_connections->table[i].id;
		iph->src() = parent->myaddr_;

		hgpsh->init();
		hgpsh->type_ = LOC_NOTIFICATION;
		hgpsh->dst_loc_ = live_connections->table[i].loc;
		hgpsh->src_loc_ = myloc;

		cmnh->ptype() = PT_HGPS;
		cmnh->size() = hgpsh->size();

		router->sendOutPktWithLocation(pkt,0.0,false);
	    }
	}
	curr_loc = myloc;
    }
 
}

void
LocServer::logTableSize()
{
    parent->trace("HGPS LocTable _%d_ %d",parent->myaddr_,table->printTableNum());
}

/* this function sends out a broadcast location update packet to 
   the node's old grid (which_loc) */
void
LocServer::updateFPGrid(location which_loc)
{
    Packet *pkt;
    hdr_hgps *hgpsh;
    hdr_cmn *cmnh;
    hdr_ip *iph;

    pkt = parent->allocpkt();
    hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_); 
    iph =  (hdr_ip*)pkt->access(parent->off_ip_); 
    cmnh =  (hdr_cmn*)pkt->access(parent->off_cmn_);

    iph->dport_ = RT_PORT;
    iph->dst() = 0; //i don't care about specific recipient
    iph->src() = parent->myaddr_;

    hgpsh->init();
    hgpsh->type_ = LOC_UPDATE_FP;
    hgpsh->src_loc_ = myloc;
    hgpsh->dst_loc_ = which_loc;

    hgpsh->update_loc_ = which_loc; //the area that the update takes effect
    
    /* -jinyang don't set tmeout interval, 
       just keep all the FPs until they overflow (karger)
       so this hgpsh->timeout_inteval_ is essentially not used*/
    hgpsh->timeout_inteval_ = max_mask *((double)parent->update_distance_/getSpeed()) + 1.5 ;
    hgpsh->timeout_inteval_ = (hgpsh->timeout_inteval_* 3/2);

    /* set up the fp pointer info */
    hgpsh->fp_ids_[0] = parent->myaddr_;
    hgpsh->fp_locs_[0] = myloc;
    hgpsh->fp_lastupdate_time_[0] = Scheduler::instance().clock();
    hgpsh->fp_num_ = 1;
    hgpsh->fp_old_loc = which_loc;

    cmnh->ptype() = PT_HGPS;
    cmnh->size() = hgpsh->size();
   
    /* instead of sending this FP pointer geographically to anybody
       in the grid, I could just broadcast it and hopefull(very likely)
       somebody in that grid will pick my FP up
    */
    cmnh->next_hop() = MAC_BROADCAST;
    iph->dst() = IP_BROADCAST;
    Scheduler::instance().schedule(parent->ll,pkt,0);
}

/* for triggered update, there is a speculative timeout value in 
   location update. In order to prevent correct location entry
   from expiring prematurely, this function is invoked to
   correct speculative timeout value when needed by sending
   out new location updates*/
void
LocServer::checkExpiringUpdates()
{
    int stop;
    double min_x,max_x,min_y,max_y;
    location tmp_loc,dst_loc;
    double speedX, speedY,speedZ;
    double dist_x, dist_y,dist_z;
    double outdate_x,outdate_y,timeout_int;
    int x, y;

    Packet *pkt;
    hdr_hgps *hgpsh;
    hdr_cmn *cmnh;
    hdr_ip *iph;

    double now = Scheduler::instance().clock();

    mynode->getVelo(&speedX,&speedY,&speedZ);
    mynode->getLoc(&dist_x,&dist_y,&dist_z);

    for (stop=2;stop<=max_mask;stop+=2) {
	
	if (now > expiration_times[stop/2-1]) {
	    
	    //I need to send out "new" updates
	    tmp_loc.loc = myloc.loc;
	    tmp_loc.mask = BIGGEST_MASK - stop + 2;

	    getXY(min_x,max_x,min_y,max_y,tmp_loc);
	    outdate_x = getOutofRangeTime(min_x,max_x,speedX,dist_x);
	    outdate_y = getOutofRangeTime(min_y,max_y,speedY,dist_y);

	    if (outdate_x < outdate_y) {
		timeout_int = outdate_x;
	    }else{
		timeout_int = outdate_y;
	    }

	    if (timeout_int < TRIGGER_PERIOD) {
		timeout_int = TRIGGER_PERIOD;
	    }
	    
	    /* store the expiration time, inflate it a bit*/
	    expiration_times[stop/2-1] = now + timeout_int + 2;
	
	    /* inflate the timeout value a bit*/
	    timeout_int = timeout_int + 4 ;

	    for (int i=0;i<4;i++) {
		if ((i << (stop-2)) == (myloc.loc & (0x3 << (stop-2)))) {
	        }else{
		
		    x = myloc.loc & (0xffffffff << stop);
		    y = i<<(stop-2);

		    dst_loc.loc = x+y;
		    dst_loc.mask = myloc.mask - stop +2;
		    assert(!inSameGrid(myloc,dst_loc));

		    pkt = parent->allocpkt();
		    hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_); 
		    iph =  (hdr_ip*)pkt->access(parent->off_ip_); 
		    cmnh =  (hdr_cmn*)pkt->access(parent->off_cmn_);

		    iph->dport_ = RT_PORT;
		    iph->dst() = 0; //i don't care about specific recipient
		    iph->src() = parent->myaddr_;

		    hgpsh->init();
		    hgpsh->src_loc_ = myloc;
		    hgpsh->dst_loc_ = dst_loc;

		    hgpsh->update_loc_ = dst_loc; //the area that the update takes effect
		    hgpsh->timeout_inteval_ = timeout_int;
		    hgpsh->type_ = LOC_UPDATE;

		    cmnh->ptype() = PT_HGPS;
		    cmnh->size() = hgpsh->size();

		    router->sendOutPktWithLocation(pkt,jitter(1,1),false);
		
		}
	
	    }

	}
    }
}

void
LocServer::updateLoc(int level) 
{
    int stop = 2;
    location dst_loc;

    Packet *pkt;
    hdr_hgps *hgpsh;
    hdr_cmn *cmnh;
    hdr_ip *iph;

    int x,y;
    double timeout_int,bigtimeout_int;
    double speedX, speedY,speedZ;
    double outdate_x,outdate_y;
    double min_x,max_x,min_y,max_y;
    double dist_x, dist_y,dist_z;
    location tmp_loc = myloc;	

    double now = Scheduler::instance().clock();

    mynode->getVelo(&speedX,&speedY,&speedZ);
    mynode->getLoc(&dist_x,&dist_y,&dist_z);

    
    while (stop<=level) {
	//determine when will i do the next triggered update
	tmp_loc.mask = BIGGEST_MASK - stop + 2;
	getXY(min_x,max_x,min_y,max_y,tmp_loc);
	outdate_x = getOutofRangeTime(min_x,max_x,speedX,dist_x);
	outdate_y = getOutofRangeTime(min_y,max_y,speedY,dist_y);

#ifdef TRIGGER_UPDATE_ONLY
	if (outdate_x < outdate_y) {
	    timeout_int = outdate_x;
	}else{
	    timeout_int = outdate_y;
	}

	if (timeout_int < TRIGGER_PERIOD) {
	    timeout_int = TRIGGER_PERIOD;
	}

	/* store the expiration time*/
	expiration_times[stop/2-1] = now + 2*timeout_int + 2;

	/* inflate the timeout value a bit*/
	timeout_int = timeout_int + 4;
#endif
	for (int i=0;i<4;i++) {
	    if ((i << (stop-2)) == (myloc.loc & (0x3 << (stop-2)))) {
	    /*
	    if (i == ((myloc.loc << (MAX_MASK-stop))>> (MAX_MASK-stop+2)) ) {
	    */
	    }else{
		x = myloc.loc & (0xffffffff << stop);
		y = i<<(stop-2);
		dst_loc.loc = x+y;
		dst_loc.mask = myloc.mask - stop +2;
		assert(!inSameGrid(dst_loc,myloc));
		pkt = parent->allocpkt();

		hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_); 
		iph =  (hdr_ip*)pkt->access(parent->off_ip_); 
		cmnh =  (hdr_cmn*)pkt->access(parent->off_cmn_);

		iph->dport_ = RT_PORT;
		iph->dst() = 0; //i don't care about specific recipient
		iph->src() = parent->myaddr_;

		hgpsh->init();
		hgpsh->src_loc_ = myloc;
		hgpsh->dst_loc_ = dst_loc;

		hgpsh->update_loc_ = dst_loc; //the area that the update takes effect
#ifdef TRIGGER_UPDATE_ONLY
		hgpsh->timeout_inteval_ = timeout_int;
#else
		hgpsh->timeout_inteval_ = 1.5 * pow(2, stop/2) * ((double)parent->update_distance_/getSpeed()) ;
#endif

		hgpsh->type_ = LOC_UPDATE;

		cmnh->ptype() = PT_HGPS;
		cmnh->size() = hgpsh->size();

		router->sendOutPktWithLocation(pkt,2*(stop-2)+jitter(1,1),false);
	    }
	}
	stop += 2;
    }

}

void
LocServer::addUpdateLiveConn(nsaddr_t id, location loc)
{
    assert(loc.mask!=0);
    live_connections->addUpdateEntry(id,loc,LIVE_CONN_TIMEOUT);
}

void
LocServer::purgeLocCache(nsaddr_t id)
{
    loc_cache->deleteEntry(id);
}

void
LocServer::updateLocCache(nsaddr_t id,location loc, double timeout_int)
{
    loc_cache->addUpdateEntry(id,loc,timeout_int);
}

LocTable::LocTable(int s, bool c)
{
    size = s;
    chopoff = c;

    table = new loc_entry[size];
    for (int i=0;i<size;i++) {
	table[i].lastupdate_time = -99999.9;
	table[i].timeout = -99999.9;
	table[i].id = -1;
    }
    used_size = 0;
}

LocTable::~LocTable()
{
    delete[] table;
}

void 
LocTable::addUpdateEntry(nsaddr_t id, location loc, double timeout)
{
    double now = Scheduler::instance().clock();
    int victim = -1;
    int j;

    for (int i = 0; i<size; i++) {
	j = (id+i) % size; //a naive hash
	if (id == table[j].id) {
	    table[j].loc = loc;
	    table[j].lastupdate_time = now;
	    table[j].timeout = timeout+now;
	    return;
	}else if ((victim == -1) && (now > table[j].timeout)) {
	    victim = j;
	}
    }
    
    if (victim == -1) {

	if (!chopoff) {
	    printf("no victim in loctable(%.2f) add entry!\n",timeout);
	    abort();
	}

    }else{
	if ((victim+1) > used_size) {
	    used_size = victim+1;
	}

	table[victim].lastupdate_time = now;
	table[victim].id = id;
	table[victim].loc = loc;
	table[victim].timeout = now+timeout;
    }

}    

void
LocTable::deleteEntry(nsaddr_t id)
{
    int j;
    for (int i=0;i<size;i++) {
	j = (id+i) % size;
	if (table[j].id == id) {
	    table[j].id = -1;
	    table[j].lastupdate_time = -99999.0;
	    table[j].timeout = -99999.0;
	    return;
	}
    }
	    
}

/* clean up all the location table */
void
LocTable::cleanTable()
{
    for (int i=0;i<size;i++) {
	table[i].id = -1;
	table[i].lastupdate_time = -9999.0;
	table[i].timeout = -9999.0;
    }
    used_size = 0;
}

loc_entry *
LocTable::getEntry(nsaddr_t id, double now)
{
    int j;
    for (int i=0;i<size;i++) {
	j = (id +i)% size;
	if ((table[j].id == id) && (now < table[j].timeout)) {
	    return &table[j];
	}
    }
    return NULL;
}

bool
LocTable::findClosestinRT(nsaddr_t myid,location myloc, nsaddr_t dst, struct route_entry *nt, loc_entry &entry, bool in_same_grid)
{

    int closest_dist = myid - dst;
    int dist;
    double now = Scheduler::instance().clock();

    entry.id = 0;

    assert(myid!=dst);

    for (int i=0;i<MAX_RT_ENTRIES;i++) {
	if (((now - nt[i].lastupdate_time)<NEIGHBOR_TIMEOUT_INTEVAL) && ((!in_same_grid) ||(nt[i].loc.loc == myloc.loc))) { 
	    dist = nt[i].dst- dst;
	    if (dist == 0) {
		entry.id = dst;
		entry.loc = nt[i].loc;
		entry.lastupdate_time = nt[i].lastupdate_time;
		return true;
	    }

	    /*
	    if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
		*/
	      if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)) {
		closest_dist = dist;
		entry.id = nt[i].dst;
		entry.loc =nt[i].loc;
		entry.lastupdate_time = nt[i].lastupdate_time;
	    }
	}
    }
    if (entry.id == 0) {
	return false;
    }else{
	return true;
    }
}

/* find the closest node in location entry database 
   in ID space to dst. If such a node is found and is closer than me, return true,
   otherwise return false */
bool
LocTable::findClosest(nsaddr_t myid,location myloc,nsaddr_t dst,struct route_entry *nt,loc_entry &entry, double lastupdate_time)
{
    int closest_dist = myid - dst;
    int dist;

    double now = Scheduler::instance().clock();

    entry.id = 0;

    assert(myid != dst);

    for (int i=0;i<MAX_RT_ENTRIES;i++) {
	if (((now - nt[i].lastupdate_time)<NEIGHBOR_TIMEOUT_INTEVAL) && (nt[i].loc.loc == myloc.loc)) {
	    dist = nt[i].dst- dst;
	    if ((dist == 0) && (nt[i].lastupdate_time > lastupdate_time)) {
		entry.id = dst;
		entry.loc = nt[i].loc;
		entry.lastupdate_time = nt[i].lastupdate_time;
		return true;
	    }
	    /*
	    if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
	    */
	    if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)) {
		closest_dist = dist;
		entry.id = nt[i].dst;
		entry.loc =nt[i].loc;
		entry.lastupdate_time = nt[i].lastupdate_time;
	    }
	}
    }

    for (int i=0;i<used_size;i++) {
	if ((table[i].id == -1) || (now > table[i].timeout)) {
	}else{
	    dist = table[i].id - dst;

	    if ((dist == 0) && (table[i].lastupdate_time > lastupdate_time)) {
		entry = table[i];
		return true;
	    }
/*
	    if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
*/
 	    if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)) {

		closest_dist = dist;
		entry = table[i];
	    }
	}
    }

    if (entry.id == 0) {
	return false;
    }else{
	return true;
    }
}

/* find the closest node in the specified grid in location entry database 
   in ID space to dst. If such a node is found and is closer than me, return true,
   otherwise return false */
bool
LocTable::findClosestInGrid(nsaddr_t myid,location myloc,nsaddr_t dst, location dst_loc,struct route_entry *nt,loc_entry &entry, double lastupdate_time)
{
    int closest_dist = myid - dst;
    int dist;

    double now = Scheduler::instance().clock();

    entry.id = 0;

    assert(myid != dst);

    for (int i=0;i<MAX_RT_ENTRIES;i++) {
	if (((now - nt[i].lastupdate_time)<NEIGHBOR_TIMEOUT_INTEVAL) && (nt[i].loc.loc == myloc.loc) && inSameGrid(nt[i].loc,dst_loc)) {
	    dist = nt[i].dst- dst;
	    if ((dist == 0) && (nt[i].lastupdate_time > lastupdate_time)) {
		entry.id = dst;
		entry.loc = nt[i].loc;
		entry.lastupdate_time = nt[i].lastupdate_time;
		return true;
	    }
	    /*
	    if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
	    */
	    if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)) {
		closest_dist = dist;
		entry.id = nt[i].dst;
		entry.loc =nt[i].loc;
		entry.lastupdate_time = nt[i].lastupdate_time;
	    }
	}
    }

    for (int i=0;i<used_size;i++) {
	if ((table[i].id == -1) || (now > table[i].timeout) || (!inSameGrid(table[i].loc,dst_loc))) {
	}else{
	    dist = table[i].id - dst;

	    if ((dist == 0) && (table[i].lastupdate_time > lastupdate_time)) {
		entry = table[i];
		return true;
	    }
/*
	    if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
*/
 	    if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)) {

		closest_dist = dist;
		entry = table[i];
	    }
	}
    }

    if (entry.id == 0) {
	return false;
    }else{
	return true;
    }
}


/* find the closest node in ID space to dst but not dst node
   this is for the node handling location update to check if itself is
   indeed the node to be updated, despite the fact that there might 
   already be location table entry for the dst node*/
void
LocTable::findClosestExceptDst(nsaddr_t myid, location myloc,nsaddr_t dst,struct route_entry *nt,location update_loc,loc_entry &entry)
{
    int closest_dist; 
    int dist;

    double now = Scheduler::instance().clock();


    if (myid == dst) {
	return;
    }

    /* if the closest entry has been initialized to some node ID, compare it with
       myid and take a closer one as the starting ID */
    if (entry.id!=0) {
	closest_dist = entry.id - dst;
	dist = myid - dst;
	/*
	if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
	*/
	if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)) {
		closest_dist = dist;
		entry.id = myid; 
       }

    }else {
	closest_dist = myid - dst;
	entry.id = myid;
    }

    assert(closest_dist != 0);
    /* part of the location table database is in one's neighbor table,
       so check neighbor table first*/
    for (int i=0;i<MAX_RT_ENTRIES;i++) {

	if (((now - nt[i].lastupdate_time)<NEIGHBOR_TIMEOUT_INTEVAL) && (inSameGrid(nt[i].loc,update_loc))) {
	 
	    dist = nt[i].dst- dst;

	    if (dist == 0) {  //ignore direct dst location entry
		continue;
	    }
	    /*
	    if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
	    */
	    if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)){
		closest_dist = dist;
		entry.id = nt[i].dst;
		entry.loc = nt[i].loc;
	    }
	}
    }

    /* check location table */
    for (int i=0;i<used_size;i++) {
	if ((table[i].id == -1) || (now > table[i].timeout)) {
	}else{
	    if (!inSameGrid(table[i].loc,update_loc)) {
		continue;
	    }

	    dist = table[i].id - dst;

	    if (dist == 0) { //ignore direct dst location entry
		continue;
	    }
	    /*
	    if (((closest_dist<0) && (dist<0) && (dist>closest_dist))
		|| ((closest_dist>0) && (dist<0))
		|| ((closest_dist>0) && (dist>0) && (dist<closest_dist))){
	    */
	    if (((closest_dist>0) && (dist>0) && (dist<closest_dist))
		|| (closest_dist<0) && (dist>0)
		|| (closest_dist<0) && (dist<0) && (dist<closest_dist)){
		closest_dist = dist;
		entry = table[i];
	    }
	}
    }

    assert(entry.id != 0 );
    return;
}

int
LocTable::printTableNum()
{
    double now = Scheduler::instance().clock();
    int counter = 0;
    for (int i=0;i<size;i++) {
	if ((table[i].id == -1) || (now > table[i].timeout)) {
	}else{
	    counter++;
	}
    }
    return counter;
}
